from datetime import datetime
from json import loads
from operator import itemgetter
from pathlib import Path
from sqlite3 import Connection
from sqlite3 import DatabaseError
from sqlite3 import OperationalError
from sqlite3 import connect
from typing import Callable
from typing import Collection
from typing import Optional
from typing import Union

__all__ = [
    "compare_versions",
    "update_database",
]


# noinspection SqlResolve,SqlNoDataSourceInspection
def get_version(conn: Connection) -> str:
    try:
        # Database version 3.0.0 and above
        return conn.execute("SELECT SVALUE FROM SETTINGS WHERE SETTING = 'VERSION'").fetchone()[0]
    except OperationalError:
        return ""


def compare_versions(a: str, b: str) -> int:
    a_split = list(map(int, a.split("-", 1)[0].split(".")))
    b_split = list(map(int, b.split("-", 1)[0].split(".")))
    a_split.extend([0] * (3 - len(a_split)))
    b_split.extend([0] * (3 - len(b_split)))

    for a_, b_ in zip(a_split, b_split):
        if a_ > b_:
            return 1
        elif a_ < b_:
            return -1
    return 0


def insert(conn: Connection, table: str, keys: Collection[str], values: Collection[Union[int, str]],
           replace: bool = True):
    conn.execute(
        f"""INSERT OR {"REPLACE" if replace else "IGNORE"} INTO {table}
        ({",".join(keys)})
        VALUES ({",".join(["?"] * len(values))})""",
        values
    )


def count(conn: Connection, table: str) -> int:
    return conn.execute(f"SELECT COUNT(*) FROM {table}").fetchall()[0][0]


def database_path(conn: Connection) -> Path | None:
    name: str
    filename: Optional[str]
    for _, name, filename in conn.execute("PRAGMA database_list"):
        if name == "main" and filename is not None:
            return Path(filename)

    return None


# noinspection SqlResolve,SqlNoDataSourceInspection,DuplicatedCode
def make_database_5(conn: Connection) -> Connection:
    conn.execute("""create table USERS
    (USERNAME text unique not null check (length(USERNAME) > 0),
    FOLDERS text not null,
    USERPAGE text not null,
    primary key (USERNAME));""")

    conn.execute("""create table SUBMISSIONS
    (ID integer unique not null check (ID > 0),
    AUTHOR text not null check (length(AUTHOR) > 0),
    TITLE text not null,
    DATE datetime not null,
    DESCRIPTION text not null,
    TAGS text not null,
    CATEGORY text not null,
    SPECIES text not null,
    GENDER text not null,
    RATING text not null,
    TYPE text not null check (TYPE in ('image', 'music', 'text', 'flash')),
    FILEURL text not null,
    FILEEXT text not null,
    FILESAVED integer not null check (FILESAVED in (0, 1, 2, 3)),
    FAVORITE text not null,
    MENTIONS text not null,
    FOLDER text not null check (FOLDER in ('gallery', 'scraps')),
    USERUPDATE text not null check (USERUPDATE in (0, 1)),
    primary key (ID));""")

    conn.execute("""create table JOURNALS
    (ID integer unique not null check (ID > 0),
    AUTHOR text not null check (length(AUTHOR) > 0),
    TITLE text not null,
    DATE datetime not null,
    CONTENT text not null,
    MENTIONS text not null,
    USERUPDATE integer not null check (USERUPDATE in (0, 1)),
    primary key (ID))""")

    conn.execute("""create table SETTINGS
    (SETTING text unique not null check (length(SETTING) > 0),
    SVALUE text check (SVALUE == null or length(SVALUE) > 0),
    primary key (SETTING));""")

    conn.execute("insert or ignore into SETTINGS (SETTING, SVALUE) values (?, ?)", ["FILESFOLDER", "FA.files"])
    conn.execute("insert or ignore into SETTINGS (SETTING, SVALUE) values (?, ?)", ["VERSION", "5.0.0"])

    conn.execute("""create table HISTORY
    (TIME datetime unique not null,
    EVENT text not null,
    primary key (TIME));""")

    conn.commit()

    return conn


# noinspection SqlResolve,SqlNoDataSourceInspection,DuplicatedCode
def make_database_5_1(conn: Connection) -> Connection:
    conn.execute("""create table USERS
    (USERNAME text unique not null check (length(USERNAME) > 0),
    FOLDERS text not null,
    ACTIVE boolean not null,
    USERPAGE text not null,
    primary key (USERNAME));""")

    conn.execute("""create table SUBMISSIONS
    (ID integer unique not null check (ID > 0),
    AUTHOR text not null check (length(AUTHOR) > 0),
    TITLE text not null,
    DATE datetime not null,
    DESCRIPTION text not null,
    TAGS text not null,
    CATEGORY text not null,
    SPECIES text not null,
    GENDER text not null,
    RATING text not null,
    TYPE text not null check (TYPE in ('image', 'music', 'text', 'flash')),
    FILEURL text not null,
    FILEEXT text not null,
    FILESAVED integer not null check (FILESAVED in (0, 1, 2, 3)),
    FAVORITE text not null,
    MENTIONS text not null,
    FOLDER text not null check (FOLDER in ('gallery', 'scraps')),
    USERUPDATE boolean not null,
    primary key (ID));""")

    conn.execute("""create table JOURNALS
    (ID integer unique not null check (ID > 0),
    AUTHOR text not null check (length(AUTHOR) > 0),
    TITLE text not null,
    DATE datetime not null,
    CONTENT text not null,
    MENTIONS text not null,
    USERUPDATE integer not null check (USERUPDATE in (0, 1)),
    primary key (ID))""")

    conn.execute("""create table SETTINGS
    (SETTING text unique not null check (length(SETTING) > 0),
    SVALUE text check (SVALUE == null or length(SVALUE) > 0),
    primary key (SETTING));""")

    conn.execute("insert or ignore into SETTINGS (SETTING, SVALUE) values (?, ?)", ["FILESFOLDER", "FA.files"])
    conn.execute("insert or ignore into SETTINGS (SETTING, SVALUE) values (?, ?)", ["VERSION", "5.1.0"])

    conn.execute("""create table HISTORY
    (TIME datetime unique not null,
    EVENT text not null,
    primary key (TIME));""")

    conn.commit()

    return conn


# noinspection SqlResolve,SqlNoDataSourceInspection,DuplicatedCode
def make_database_5_2(conn: Connection) -> Connection:
    conn.execute("""create table USERS
    (USERNAME text unique not null check (length(USERNAME) > 0),
    FOLDERS text not null,
    ACTIVE boolean not null,
    USERPAGE text not null,
    primary key (USERNAME));""")

    conn.execute("""create table SUBMISSIONS
    (ID integer unique not null check (ID > 0),
    AUTHOR text not null check (length(AUTHOR) > 0),
    TITLE text not null,
    DATE datetime not null,
    DESCRIPTION text not null,
    TAGS text not null,
    CATEGORY text not null,
    SPECIES text not null,
    GENDER text not null,
    RATING text not null,
    TYPE text not null check (TYPE in ('image', 'music', 'text', 'flash')),
    FILEURL text not null,
    FILEEXT text not null,
    FILESAVED integer not null check (FILESAVED in (0, 1, 2, 3)),
    FAVORITE text not null,
    MENTIONS text not null,
    FOLDER text not null check (FOLDER in ('gallery', 'scraps')),
    USERUPDATE boolean not null,
    primary key (ID));""")

    conn.execute("""create table JOURNALS
    (ID integer unique not null check (ID > 0),
    AUTHOR text not null check (length(AUTHOR) > 0),
    TITLE text not null,
    DATE datetime not null,
    CONTENT text not null,
    MENTIONS text not null,
    USERUPDATE integer not null check (USERUPDATE in (0, 1)),
    primary key (ID))""")

    conn.execute("""create table COMMENTS
    (ID integer unique not null check (ID > 0),
    PARENT_TABLE text not null check (PARENT_TABLE in ('SUBMISSIONS', 'JOURNALS')),
    PARENT_ID integer not null check (PARENT_ID > 0),
    REPLY_TO integer check (REPLY_TO == null or REPLY_TO > 0),
    AUTHOR text not null check (length(AUTHOR) > 0),
    DATE datetime not null,
    TEXT text not null,
    primary key (ID))""")

    conn.execute("""create table SETTINGS
    (SETTING text unique not null check (length(SETTING) > 0),
    SVALUE text check (SVALUE == null or length(SVALUE) > 0),
    primary key (SETTING));""")

    conn.execute("insert or ignore into SETTINGS (SETTING, SVALUE) values (?, ?)", ["FILESFOLDER", "FA.files"])
    conn.execute("insert or ignore into SETTINGS (SETTING, SVALUE) values (?, ?)", ["VERSION", "5.2.0"])

    conn.execute("""create table HISTORY
    (TIME datetime unique not null,
    EVENT text not null,
    primary key (TIME));""")

    conn.commit()

    return conn


# noinspection SqlResolve,SqlNoDataSourceInspection,DuplicatedCode
def make_database_5_2_2(conn: Connection) -> Connection:
    conn.execute("""create table USERS
    (USERNAME text unique not null check (length(USERNAME) > 0),
    FOLDERS text not null,
    ACTIVE boolean not null,
    USERPAGE text not null,
    primary key (USERNAME));""")

    conn.execute("""create table SUBMISSIONS
    (ID integer unique not null check (ID > 0),
    AUTHOR text not null check (length(AUTHOR) > 0),
    TITLE text not null,
    DATE datetime not null,
    DESCRIPTION text not null,
    TAGS text not null,
    CATEGORY text not null,
    SPECIES text not null,
    GENDER text not null,
    RATING text not null,
    TYPE text not null check (TYPE in ('image', 'music', 'text', 'flash')),
    FILEURL text not null,
    FILEEXT text not null,
    FILESAVED integer not null check (FILESAVED in (0, 1, 2, 3)),
    FAVORITE text not null,
    MENTIONS text not null,
    FOLDER text not null check (FOLDER in ('gallery', 'scraps')),
    USERUPDATE boolean not null,
    primary key (ID));""")

    conn.execute("""create table JOURNALS
    (ID integer unique not null check (ID > 0),
    AUTHOR text not null check (length(AUTHOR) > 0),
    TITLE text not null,
    DATE datetime not null,
    CONTENT text not null,
    MENTIONS text not null,
    USERUPDATE integer not null check (USERUPDATE in (0, 1)),
    primary key (ID))""")

    conn.execute("""create table COMMENTS
    (ID integer not null check (ID > 0),
    PARENT_TABLE text not null check (PARENT_TABLE in ('SUBMISSIONS', 'JOURNALS')),
    PARENT_ID integer not null check (PARENT_ID > 0),
    REPLY_TO integer check (REPLY_TO == null or REPLY_TO > 0),
    AUTHOR text not null check (length(AUTHOR) > 0),
    DATE datetime not null,
    TEXT text not null,
    primary key (ID, PARENT_TABLE, PARENT_ID))""")

    conn.execute("""create table SETTINGS
    (SETTING text unique not null check (length(SETTING) > 0),
    SVALUE text check (SVALUE == null or length(SVALUE) > 0),
    primary key (SETTING));""")

    conn.execute("insert or ignore into SETTINGS (SETTING, SVALUE) values (?, ?)", ["FILESFOLDER", "FA.files"])
    conn.execute("insert or ignore into SETTINGS (SETTING, SVALUE) values (?, ?)", ["VERSION", "5.2.2"])

    conn.execute("""create table HISTORY
    (TIME datetime unique not null,
    EVENT text not null,
    primary key (TIME));""")

    conn.commit()

    return conn


# noinspection SqlResolve,SqlNoDataSourceInspection,DuplicatedCode
def make_database_5_3(conn: Connection) -> Connection:
    conn.execute("""create table USERS
    (USERNAME text unique not null check (length(USERNAME) > 0),
    FOLDERS text not null,
    ACTIVE boolean not null,
    USERPAGE text not null,
    primary key (USERNAME));""")

    conn.execute("""create table SUBMISSIONS
    (ID integer unique not null check (ID > 0),
    AUTHOR text not null check (length(AUTHOR) > 0),
    TITLE text not null,
    DATE datetime not null,
    DESCRIPTION text not null,
    TAGS text not null,
    CATEGORY text not null,
    SPECIES text not null,
    GENDER text not null,
    RATING text not null,
    TYPE text not null check (TYPE in ('image', 'music', 'text', 'flash')),
    FILEURL text not null,
    FILEEXT text not null,
    FILESAVED integer not null check (FILESAVED in (0, 1, 2, 3, 4, 5, 6, 7)),
    FAVORITE text not null,
    MENTIONS text not null,
    FOLDER text not null check (FOLDER in ('gallery', 'scraps')),
    USERUPDATE boolean not null,
    primary key (ID));""")

    conn.execute("""create table JOURNALS
    (ID integer unique not null check (ID > 0),
    AUTHOR text not null check (length(AUTHOR) > 0),
    TITLE text not null,
    DATE datetime not null,
    CONTENT text not null,
    MENTIONS text not null,
    USERUPDATE integer not null check (USERUPDATE in (0, 1)),
    primary key (ID))""")

    conn.execute("""create table COMMENTS
    (ID integer not null check (ID > 0),
    PARENT_TABLE text not null check (PARENT_TABLE in ('SUBMISSIONS', 'JOURNALS')),
    PARENT_ID integer not null check (PARENT_ID > 0),
    REPLY_TO integer check (REPLY_TO == null or REPLY_TO > 0),
    AUTHOR text not null check (length(AUTHOR) > 0),
    DATE datetime not null,
    TEXT text not null,
    primary key (ID, PARENT_TABLE, PARENT_ID))""")

    conn.execute("""create table SETTINGS
    (SETTING text unique not null check (length(SETTING) > 0),
    SVALUE text check (SVALUE == null or length(SVALUE) > 0),
    primary key (SETTING));""")

    conn.execute("insert or ignore into SETTINGS (SETTING, SVALUE) values (?, ?)", ["FILESFOLDER", "FA.files"])
    conn.execute("insert or ignore into SETTINGS (SETTING, SVALUE) values (?, ?)", ["BACKUPFOLDER", "FA.backup"])
    conn.execute("insert or ignore into SETTINGS (SETTING, SVALUE) values (?, ?)", ["VERSION", "5.3.0"])

    conn.execute("""create table HISTORY
    (TIME datetime unique not null,
    EVENT text not null,
    primary key (TIME));""")

    conn.commit()

    return conn


# noinspection SqlResolve,SqlNoDataSourceInspection,DuplicatedCode
def update_5_0(conn: Connection, _db_path: Path, db_new_path: Path):
    make_database_5(connect(db_new_path)).close()
    conn.execute("attach database ? as db_new", [str(db_new_path)])
    conn.execute("insert into db_new.USERS (USERNAME, FOLDERS, USERPAGE)"
                 "select USERNAME, FOLDERS, '' from USERS")
    conn.execute("insert into db_new.SUBMISSIONS (ID,AUTHOR,TITLE,DATE,DESCRIPTION,TAGS,CATEGORY,SPECIES,GENDER,RATING,"
                 "TYPE,FILEURL,FILEEXT,FAVORITE,MENTIONS,FOLDER,USERUPDATE,FILESAVED)"
                 "select ID,AUTHOR,TITLE,DATE,DESCRIPTION,TAGS,CATEGORY,SPECIES,GENDER,RATING,TYPE,FILEURL,FILEEXT,"
                 "FAVORITE,MENTIONS,FOLDER,USERUPDATE,(((FILESAVED >= 10) * 2 ) + (FILESAVED % 10 == 1))"
                 "from SUBMISSIONS")
    conn.execute("insert into db_new.JOURNALS select * from JOURNALS")
    conn.execute("insert or replace into db_new.SETTINGS select * from SETTINGS"
                 " where SETTING not in ('VERSION', 'HISTORY')")
    history: list[tuple[float, str]] = loads(
        conn.execute("select SVALUE from SETTINGS where SETTING = 'HISTORY'").fetchone()[0] or "[]")
    for time, event in sorted(history, key=itemgetter(0)):
        conn.execute("insert into db_new.HISTORY (TIME, EVENT) values (?, ?)",
                     (datetime.fromtimestamp(time).strftime("%Y-%m-%dT%H:%M:%S.%f"), event))


# noinspection SqlResolve,SqlNoDataSourceInspection,DuplicatedCode
def update_5_0_10(conn: Connection, _db_path: Path, db_new_path: Path) -> list[str]:
    messages: list[str] = []
    make_database_5(connect(db_new_path)).close()
    conn.execute("attach database ? as db_new", [str(db_new_path)])
    conn.execute("insert into db_new.USERS select* from USERS")
    conn.execute("insert into db_new.SUBMISSIONS select * from SUBMISSIONS")
    conn.execute("insert into db_new.JOURNALS select * from JOURNALS")
    conn.execute("insert into db_new.HISTORY select * from HISTORY")
    conn.execute("insert or replace into db_new.SETTINGS select * from SETTINGS"
                 " where SETTING != 'VERSION'")
    conn.execute("update db_new.SETTINGS set SVALUE = '5.0.10' where SETTING = 'VERSION'")
    users_favorites: list[str] = [
        u for [u] in conn.execute("select USERNAME from USERS where FOLDERS like '%favorites%'").fetchall()]
    modified: int = 0
    for i, fs_raw in conn.execute("select ID, FAVORITE from SUBMISSIONS where FAVORITE like '%|_|%'").fetchall():
        fs: list[str] = list(filter(bool, fs_raw.split("|")))
        fs_filtered = list(filter(users_favorites.__contains__, fs))
        if fs != fs_filtered:
            conn.execute("update db_new.SUBMISSIONS set FAVORITE = ? where ID = ?",
                         ("".join(f"|{f}|" for f in fs_filtered), i))
            modified += 1
    if modified:
        messages.append(f"{modified} submissions modified.")
    return messages


# noinspection SqlResolve,SqlNoDataSourceInspection,DuplicatedCode
def update_5_1(conn: Connection, _db_path: Path, db_new_path: Path) -> list[str]:
    make_database_5_1(connect(db_new_path)).close()
    conn.execute("attach database ? as db_new", [str(db_new_path)])
    conn.execute("insert into db_new.USERS (USERNAME, FOLDERS, USERPAGE, ACTIVE)"
                 "select USERNAME, replace(FOLDERS, '!', ''), USERPAGE, FOLDERS not like '%!%' from USERS")
    conn.execute("insert into db_new.SUBMISSIONS select * from SUBMISSIONS")
    conn.execute("insert into db_new.JOURNALS select * from JOURNALS")
    conn.execute("insert into db_new.HISTORY select * from HISTORY")
    conn.execute("insert or replace into db_new.SETTINGS select * from SETTINGS where SETTING != 'VERSION'")
    conn.execute("update db_new.SETTINGS set SVALUE = '5.1.0' where SETTING = 'VERSION'")
    return []


# noinspection SqlResolve,SqlNoDataSourceInspection,DuplicatedCode,SqlWithoutWhere
def update_5_1_2(conn: Connection, _db_path: Path, db_new_path: Path) -> list[str]:
    make_database_5_1(connect(db_new_path)).close()
    conn.execute("attach database ? as db_new", [str(db_new_path)])
    conn.execute("insert into db_new.USERS select * from USERS")
    conn.execute("insert into db_new.SUBMISSIONS select * from SUBMISSIONS")
    conn.execute("insert into db_new.JOURNALS select * from JOURNALS")
    conn.execute("insert into db_new.HISTORY select * from HISTORY")
    conn.execute("insert or replace into db_new.SETTINGS select * from SETTINGS where SETTING != 'VERSION'")
    conn.execute("update db_new.SETTINGS set SVALUE = '5.1.2' where SETTING = 'VERSION'")
    conn.execute("""
        update db_new.SUBMISSIONS 
        set CATEGORY = replace(replace(replace(CATEGORY, '/ ', '/'), ' /', '/'), '/', ' / '),
            SPECIES = replace(replace(replace(SPECIES, '/ ', '/'), ' /', '/'), '/', ' / ')""")
    return []


# noinspection SqlResolve,SqlNoDataSourceInspection,DuplicatedCode,SqlWithoutWhere
def update_5_2(conn: Connection, _db_path: Path, db_new_path: Path) -> list[str]:
    make_database_5_2(connect(db_new_path)).close()
    conn.execute("attach database ? as db_new", [str(db_new_path)])
    conn.execute("insert into db_new.USERS select * from USERS")
    conn.execute("insert into db_new.SUBMISSIONS select * from SUBMISSIONS")
    conn.execute("insert into db_new.JOURNALS select * from JOURNALS")
    conn.execute("insert into db_new.HISTORY select * from HISTORY")
    conn.execute("insert or replace into db_new.SETTINGS select * from SETTINGS where SETTING != 'VERSION'")
    conn.execute("update db_new.SETTINGS set SVALUE = '5.2.0' where SETTING = 'VERSION'")
    return []


# noinspection SqlResolve,SqlNoDataSourceInspection,DuplicatedCode,SqlWithoutWhere
def update_5_2_2(conn: Connection, _db_path: Path, db_new_path: Path) -> list[str]:
    make_database_5_2_2(connect(db_new_path)).close()
    conn.execute("attach database ? as db_new", [str(db_new_path)])
    conn.execute("insert into db_new.USERS select * from USERS")
    conn.execute("insert into db_new.SUBMISSIONS select * from SUBMISSIONS")
    conn.execute("insert into db_new.JOURNALS select * from JOURNALS")
    conn.execute("insert into db_new.COMMENTS select * from COMMENTS")
    conn.execute("insert into db_new.HISTORY select * from HISTORY")
    conn.execute("insert or replace into db_new.SETTINGS select * from SETTINGS where SETTING != 'VERSION'")
    conn.execute("update db_new.SETTINGS set SVALUE = '5.2.2' where SETTING = 'VERSION'")
    return []


# noinspection SqlResolve,SqlNoDataSourceInspection,DuplicatedCode,SqlWithoutWhere
def update_5_3(conn: Connection, _db_path: Path, db_new_path: Path) -> list[str]:
    make_database_5_3(connect(db_new_path)).close()
    conn.execute("attach database ? as db_new", [str(db_new_path)])
    conn.execute("insert into db_new.USERS select * from USERS")
    conn.execute("insert into db_new.SUBMISSIONS select * from SUBMISSIONS")
    conn.execute("insert into db_new.JOURNALS select * from JOURNALS")
    conn.execute("insert into db_new.COMMENTS select * from COMMENTS")
    conn.execute("insert into db_new.HISTORY select * from HISTORY")
    conn.execute("insert or replace into db_new.SETTINGS select * from SETTINGS where SETTING != 'VERSION'")
    conn.execute("update db_new.SETTINGS set SVALUE = '5.3.0' where SETTING = 'VERSION'")
    conn.execute("update db_new.SUBMISSIONS set FILEURL = ('|' || FILEURL || '|')")
    conn.execute("update db_new.SUBMISSIONS set FILEEXT = ('|' || FILEEXT || '|') where FILESAVED & 2")
    conn.execute("update db_new.SUBMISSIONS set FILESAVED = FILESAVED + 4 where FILESAVED & 2")
    return []


# noinspection SqlResolve,SqlNoDataSourceInspection,DuplicatedCode,SqlWithoutWhere
def update_5_3_2(conn: Connection, _db_path: Path, db_new_path: Path) -> list[str]:
    make_database_5_3(connect(db_new_path)).close()
    conn.execute("attach database ? as db_new", [str(db_new_path)])
    conn.execute("insert into db_new.USERS select * from USERS")
    conn.execute("insert into db_new.SUBMISSIONS select * from SUBMISSIONS")
    conn.execute("insert into db_new.JOURNALS select * from JOURNALS")
    conn.execute("insert into db_new.COMMENTS select * from COMMENTS")
    conn.execute("insert into db_new.HISTORY select * from HISTORY")
    conn.execute("insert or replace into db_new.SETTINGS select * from SETTINGS where SETTING != 'VERSION'")
    conn.execute("update db_new.SETTINGS set SVALUE = '5.3.2' where SETTING = 'VERSION'")

    def tiered_path(i: int | str, depth: int = 5, width: int = 2) -> Path:
        id_str: str = str(int(i)).zfill(depth * width)
        return Path(*[id_str[n_:n_ + width] for n_ in range(0, depth * width, width)])

    files_folder: Path = Path(conn.execute("select SVALUE from SETTINGS where SETTING = 'FILESFOLDER'").fetchone()[0])
    submissions = conn.execute("""select ID, FILEEXT from db_new.SUBMISSIONS
        where FILEEXT like '%|||%' or FILEEXT like '%||' order by ID""")
    submissions_fixed: int = 0

    for [id_, exts_raw] in submissions:
        exts = exts_raw.removeprefix("|").removesuffix("|").split("||")
        folder: Path = files_folder / tiered_path(id_)
        for n, ext in [(n, e) for n, e in enumerate(exts) if "|" in e]:
            ext_new = ext.removesuffix("|")
            file = folder / f"submission{n if n else ''}.{ext}"
            if file.is_file():
                file.replace(file.with_suffix(ext_new))
            exts[n] = ext_new
        conn.execute("update db_new.SUBMISSIONS set FILEEXT = ? where ID = ?", [f"|{'|'.join(exts)}|", id_])
        submissions_fixed += 1
    return [f"{submissions_fixed} submissions extensions fixed"]


def update_wrapper(conn: Connection, update_function: Callable[[Connection, Path, Path], list[str] | None],
                   version_old: str, version_new: str) -> Connection:
    print(f"Updating {version_old} to {version_new}... ", end="", flush=True)
    db_path: Path = p if (p := database_path(conn)) else Path("FA.db")
    db_new_path: Path = db_path.with_name(f".new_{db_path.name}")
    db_new_path.unlink(missing_ok=True)
    try:
        messages: list[str] = update_function(conn, db_path, db_new_path) or []
        conn.commit()
        conn.close()
        conn = None
        orig_name: str = db_path.name
        db_path.replace(db_path := db_path.with_name(f"v{version_old.replace('.', '_')}_{db_path.name}"))
        db_new_path.replace(db_new_path := db_new_path.with_name(orig_name))
        print("Complete")
        for message in messages:
            print(f"  {message}")
        print("  Previous version moved to:", db_path, end="", flush=True)
        return connect(db_new_path)
    except BaseException as err:
        conn.close()
        conn = None
        db_new_path.unlink(missing_ok=True)
        raise err
    finally:
        if conn is not None:
            conn.commit()
            conn.close()
        print()


# noinspection SqlResolve,SqlNoDataSourceInspection
def update_patch(conn: Connection, version: str, target_version: str) -> Connection:
    print(f"Patching {version} to {target_version}... ", end="", flush=True)
    conn.execute("UPDATE SETTINGS SET SVALUE = ? WHERE SETTING = 'VERSION'", [target_version])
    conn.commit()
    print("Complete")
    return conn


def update_database(conn: Connection, version: str) -> Connection:
    if not (db_version := get_version(conn)):
        raise DatabaseError("Cannot read version from database.")
    elif (v := compare_versions(db_version, version)) == 0:
        return conn
    elif v > 0:
        raise DatabaseError("Database version is newer than program.")
    elif compare_versions(db_version, "4.19.0") < 0:
        raise DatabaseError("Update does not support versions lower than 4.19.0.")
    elif compare_versions(db_version, v := "5.0.0") < 0:
        conn = update_wrapper(conn, update_5_0, db_version, v)  # 4.19.x to 5.0.0
    elif compare_versions(db_version, v := "5.0.10") < 0:
        conn = update_wrapper(conn, update_5_0_10, db_version, v)  # 5.0.x to 5.0.10
    elif compare_versions(db_version, v := "5.1.0") < 0:
        conn = update_wrapper(conn, update_5_1, db_version, v)  # 5.0.10 to 5.1.0
    elif compare_versions(db_version, v := "5.1.2") < 0:
        conn = update_wrapper(conn, update_5_1_2, db_version, v)  # 5.1.0-5.1.1 to 5.1.2
    elif compare_versions(db_version, v := "5.2.0") < 0:
        conn = update_wrapper(conn, update_5_2, db_version, v)  # 5.1.2 to 5.2.0
    elif compare_versions(db_version, v := "5.2.2") < 0:
        conn = update_wrapper(conn, update_5_2_2, db_version, v)  # 5.2.0-5.2.1 to 5.2.2
    elif compare_versions(db_version, v := "5.3.0") < 0:
        conn = update_wrapper(conn, update_5_3, db_version, v)  # 5.2.2 to 5.3.0
    elif compare_versions(db_version, v := "5.3.2") < 0:
        conn = update_wrapper(conn, update_5_3_2, db_version, v)  # 5.3.0 to 5.3.2
    elif compare_versions(db_version, version) < 0:
        return update_patch(conn, db_version, version)  # Update to the latest patch

    return update_database(conn, version)



from datetime import datetime
from enum import Enum
from enum import EnumMeta

from .column import Column
from .column import parse_list
from .util import clean_username

__all__ = [
    "users_table",
    "submissions_table",
    "journals_table",
    "comments_table",
    "settings_table",
    "history_table",
    "UsersColumns",
    "SubmissionsColumns",
    "JournalsColumns",
    "CommentsColumns",
    "SettingsColumns",
    "HistoryColumns",
]

users_table: str = "USERS"
submissions_table: str = "SUBMISSIONS"
journals_table: str = "JOURNALS"
comments_table: str = "COMMENTS"
settings_table: str = "SETTINGS"
history_table: str = "HISTORY"


class ColumnsEnum(Enum):
    @classmethod
    def as_list(cls: EnumMeta) -> list[Column]:
        return [c.value for c in cls]


class UsersColumns(ColumnsEnum):
    USERNAME = Column("USERNAME", str, unique=True, key=True, check="length({name}) > 0", to_entry=clean_username)
    FOLDERS = Column("FOLDERS", set)
    ACTIVE = Column("ACTIVE", bool)
    USERPAGE = Column("USERPAGE", str, to_entry=str.strip)


class SubmissionsColumns(ColumnsEnum):
    ID = Column("ID", int, unique=True, key=True, check="{name} > 0")
    AUTHOR = Column("AUTHOR", str, check="length({name}) > 0")
    TITLE = Column("TITLE", str)
    DATE = Column("DATE", datetime)
    DESCRIPTION = Column("DESCRIPTION", str)
    TAGS = Column("TAGS", list)
    CATEGORY = Column("CATEGORY", str)
    SPECIES = Column("SPECIES", str)
    GENDER = Column("GENDER", str)
    RATING = Column("RATING", str)
    TYPE = Column("TYPE", str, to_entry=str.lower, check="{name} in ('image', 'music', 'text', 'flash')")
    FILEURL = Column("FILEURL", list[str], from_entry=lambda v: parse_list(v))
    FILEEXT = Column("FILEEXT", list[str], from_entry=lambda v: parse_list(v))
    FILESAVED = Column("FILESAVED", int, check="{name} in (0, 1, 2, 3, 4, 5, 6, 7)")
    FAVORITE = Column("FAVORITE", set)
    MENTIONS = Column("MENTIONS", set)
    FOLDER = Column("FOLDER", str, to_entry=str.lower, check="{name} in ('gallery', 'scraps')")
    USERUPDATE = Column("USERUPDATE", bool)


class JournalsColumns(ColumnsEnum):
    ID = Column("ID", int, unique=True, key=True, check="{name} > 0")
    AUTHOR = Column("AUTHOR", str, check="length({name}) > 0")
    TITLE = Column("TITLE", str)
    DATE = Column("DATE", datetime)
    CONTENT = Column("CONTENT", str)
    MENTIONS = Column("MENTIONS", set)
    USERUPDATE = Column("USERUPDATE", bool)


class CommentsColumns(ColumnsEnum):
    ID = Column("ID", int, key=True, check="{name} > 0")
    PARENT_TABLE = Column("PARENT_TABLE", str, key=True, check=f"{'{name}'} in ('{submissions_table}', '{journals_table}')")
    PARENT_ID = Column("PARENT_ID", int, key=True, check="{name} > 0")
    REPLY_TO = Column("REPLY_TO", int, not_null=False, check="{name} == null or {name} > 0")
    AUTHOR = Column("AUTHOR", str, check="length({name}) > 0")
    DATE = Column("DATE", datetime, to_entry=lambda v: v.strftime("%Y-%m-%dT%H:%M:%S"),
                  from_entry=lambda v: datetime.strptime(v, "%Y-%m-%dT%H:%M:%S"))
    TEXT = Column("TEXT", str)


class SettingsColumns(ColumnsEnum):
    SETTING = Column("SETTING", str, unique=True, key=True, check="length({name}) > 0")
    SVALUE = Column("SVALUE", str, not_null=False, check="{name} == null or length({name}) > 0")


class HistoryColumns(ColumnsEnum):
    TIME = Column("TIME", datetime, unique=True, key=True, to_entry=lambda v: v.strftime("%Y-%m-%dT%H:%M:%S.%f"),
                  from_entry=lambda v: datetime.strptime(v, "%Y-%m-%dT%H:%M:%S.%f"))
    EVENT = Column("EVENT", str)



from datetime import datetime
from os import PathLike
from pathlib import Path
from re import search
from shutil import copy
from shutil import copy2
from shutil import move
from sqlite3 import Connection
from sqlite3 import Cursor as SQLCursor
from sqlite3 import DatabaseError
from sqlite3 import ProgrammingError
from sqlite3 import connect
from typing import Any
from typing import Generator
from typing import Iterable
from typing import Type
from typing import TypeVar
from typing import overload

from psutil import Process

from .__version__ import __version__
from .column import Column
from .column import NoDefault
from .exceptions import VersionError
from .selector import AND
from .selector import EQ
from .selector import OR
from .selector import Selector
from .selector import selector_to_sql
from .tables import CommentsColumns
from .tables import HistoryColumns
from .tables import JournalsColumns
from .tables import SettingsColumns
from .tables import SubmissionsColumns
from .tables import UsersColumns
from .tables import comments_table
from .tables import history_table
from .tables import journals_table
from .tables import settings_table
from .tables import submissions_table
from .tables import users_table
from .types import Value
from .update import update_database
from .util import clean_username
from .util import compare_version
from .util import find_connections
from .util import guess_extension
from .util import query_to_sql
from .util import tiered_path

T = TypeVar("T")


def _copy_folder(src: Path, dest: Path):
    if src.is_dir():
        for item in src.iterdir():
            _copy_folder(item, dest / item.name)
    elif src.is_file and not dest.is_file():
        dest.parent.mkdir(parents=True, exist_ok=True)
        copy(src, dest)


def copy_cursors(db_dest: 'Database', cursors: Iterable['Cursor'], replace: bool, exist_ok: bool):
    if not cursors:
        return
    elif not db_dest.is_formatted:
        raise DatabaseError("Destination database is not formatted.")
    elif any(c.table.database.path == db_dest.path for c in cursors):
        raise DatabaseError("Cursors must not point to the destination database.")
    elif err := compare_version(version_dest := db_dest.version):
        raise VersionError("Destination database is not up to date.", *err.args)
    elif any(compare_version(c.table.database.version, version_b=version_dest) is not None for c in cursors):
        raise VersionError("Cursors must point to a database with the same version as the destination database.")
    elif any(set(c_.name for c_ in c.columns) != set(c_.name for c_ in c.table.columns) for c in cursors):
        raise DatabaseError("Cursors must contain all columns.")

    for cursor in cursors:
        cursor_db: Database = cursor.table.database
        dest_table: Table
        if cursor.table.name.lower() == db_dest.users.name.lower():
            dest_table = db_dest.users
        elif cursor.table.name.lower() == db_dest.submissions.name.lower():
            dest_table = db_dest.submissions
        elif cursor.table.name.lower() == db_dest.journals.name.lower():
            dest_table = db_dest.journals
        elif cursor.table.name.lower() == db_dest.comments.name.lower():
            dest_table = db_dest.comments
        elif cursor.table.name.lower() == db_dest.settings.name.lower():
            dest_table = db_dest.settings
        elif cursor.table.name.lower() == db_dest.history.name.lower():
            dest_table = db_dest.history
        else:
            raise DatabaseError(f"Unknown table {cursor.table.name}")
        for entry in cursor:
            if dest_table.name.lower() == db_dest.submissions.name.lower():
                fs, t = cursor_db.submissions.get_submission_files(entry[cursor.table.key.name])
                db_dest.submissions.save_submission(entry, [f.read_bytes() for f in fs or []],
                                                    t.read_bytes() if t else None,
                                                    replace=replace, exist_ok=exist_ok)
            else:
                dest_table.insert(dest_table.format_entry(entry), replace=replace, exists_ok=True)


class Cursor:
    def __init__(self, cursor: SQLCursor, columns: list[Column], table: 'Table', *, query: str = None,
                 query_values: list[Any] = None):
        self.cursor: SQLCursor = cursor
        self.columns: list[Column] = columns
        self.table: Table = table
        self.query: str | None = query
        self.query_values: list[Any] | None = query_values

    def __next__(self) -> dict[str, Value]:
        return next(self.entries)

    def __iter__(self) -> Generator[dict[str, Value], None, None]:
        return self.entries

    @property
    def entries(self) -> Generator[dict[str, Any], None, None]:
        return ({c.name: c.from_entry(v) for c, v in zip(self.columns, row, strict=True)} for row in self.cursor)

    @property
    def tuples(self) -> Generator[tuple, None, None]:
        return (tuple(c.from_entry(v) for c, v in zip(self.columns, row, strict=True)) for row in self.cursor)

    def fetchone(self):
        return next(self.entries, None)

    def fetchall(self):
        return list(self.entries)


class Table:
    def __init__(self, database: "Database", name: str, columns: Iterable[Column] = None):
        self.database: Database = database
        self.name: str = name
        self._columns: list[Column] = columns or []

    def __len__(self) -> int:
        return self.select(columns=[Column(f"count({self.key.name})", int)]).cursor.fetchone()[0]

    def __contains__(self, key: Value) -> bool:
        return bool(self[key])

    @overload
    def __getitem__(self, key: dict[str, Value] | tuple[Value] | list[Value]) -> list[dict[str, Value]]:
        ...

    @overload
    def __getitem__(self, key: Value) -> dict[str, Value] | None:
        ...

    def __getitem__(self, key: Value | dict[str, Value] | tuple[Value]
                    ) -> dict[str, Value] | list[dict[str, Value]] | None:
        if isinstance(key, dict):
            return self.select(
                {AND: [{EQ: {k: v}} for k, v in self.format_entry(key, defaults=False).items()]}).fetchall()
        elif isinstance(key, (tuple, list)):
            return self.select({OR: [{EQ: {self.key.name: self.key.to_entry(k)}} for k in key]}).fetchall()
        else:
            return self.select({EQ: {self.key.name: self.key.to_entry(key)}}).fetchone()

    def __setitem__(self, key: Value, entry: dict[str, Any]):
        self.insert(self.format_entry(entry | {self.key.name: key}), replace=True)

    def __delitem__(self, key: Value | dict[str, Value] | tuple[Value] | list[Value]) -> SQLCursor:
        if isinstance(key, dict):
            return self.delete({EQ: self.format_entry(key, defaults=False)})
        elif isinstance(key, (tuple, list)):
            return self.delete({OR: [{EQ: {self.key.name: self.key.to_entry(k)}} for k in key]})
        else:
            return self.delete({EQ: {self.key.name: self.key.to_entry(key)}})

    def __iter__(self) -> Generator[dict[str, Value], None, None]:
        return self.select().entries

    def _get_exists(self, key: Value) -> dict:
        if not (entry := self[key]):
            raise KeyError(f"Entry {self.key.name} = {key!r} does not exist in {self.name} table.")
        return entry

    def _get_columns(self) -> list[Column]:
        return [Column(name, t, not_null=bool(not_null), key=pk)
                for _, name, t, not_null, _, pk in self.database.execute(f"pragma table_info({self.name})")]

    @property
    def columns(self) -> list[Column]:
        self._columns = self._columns or self._get_columns()
        return self._columns

    @property
    def key(self) -> Column | None:
        return next((k for k in self.keys), None)

    @property
    def keys(self) -> list[Column]:
        return [c for c in self.columns if c.key]

    def get_column(self, name: str) -> Column | None:
        name = name.lower()
        return next((c for c in self.columns if c.name.lower() == name.lower()), None)

    def create_statement(self, exists_ignore: bool = False) -> str:
        elements: list[str] = ["create table"]

        if exists_ignore:
            elements.append("if not exists")

        elements.append(self.name)

        if self.columns:
            elements.append("(" + ", ".join(c.create_statement() for c in self.columns) +
                            (f", primary key ({', '.join(c.name for c in keys)})" if (keys := self.keys) else "") +
                            ")")

        return " ".join(elements)

    def create(self, exists_ignore: bool = True):
        self.database.execute(self.create_statement(exists_ignore=exists_ignore))

    def format_entry(self, entry: dict[str, Any], *, defaults: bool = True) -> dict[str, Value]:
        columns_dict: dict[str, Any] = {}
        if defaults:
            columns_dict = {c.name.upper(): c.default for c in self.columns if c.default is not NoDefault}
        entry = columns_dict | {k.upper(): v for k, v in entry.items()}
        entry = {(c := self.get_column(k)).name: c.to_entry(v) for k, v in entry.items()}
        return entry

    def insert(self, entry: dict[str, Value], *, replace: bool = False, exists_ok: bool = False):
        self.database.execute(
            f"""INSERT {'OR REPLACE' if replace else 'OR IGNORE' if exists_ok else ''} INTO {self.name}
                    ({','.join(entry.keys())}) VALUES ({','.join(['?'] * len(entry))})""",
            [v for v in entry.values()]
        )

    def select(self, query: Selector = None, columns: list[str | Column] = None, order: list[str] = None,
               limit: int = 0,
               offset: int = 0) -> Cursor:
        sql, values = selector_to_sql(query) if query else ("", None)
        return self.select_sql(sql, values, columns, order, limit, offset)

    def select_query(self, query: str, columns: list[str | Column] = None, default_field: str = None,
                     likes: list[str] = None, aliases: dict[str, str] = None, order: list[str] = None, limit: int = 0,
                     offset: int = 0) -> Cursor:
        elements, values = query_to_sql(query, default_field or self.key.name, likes, aliases)
        return self.select_sql(" ".join(elements), values, columns, order, limit, offset)

    def select_sql(self, sql: str, values: list[Any] = None, columns: list[str | Column] = None,
                   order: list[str] = None, limit: int = 0, offset: int = 0) -> Cursor:
        columns_: list[Column] = [(self.get_column(c) or Column(c, Any)) if isinstance(c, str) else c
                                  for c in columns] if columns else self.columns
        sql = " ".join(list(filter(bool, [f"SELECT {','.join(c.name for c in columns_)} FROM {self.name}",
                                          f"WHERE {sql}" if sql else None,
                                          f"ORDER BY {','.join(order)}" if order else None,
                                          f"LIMIT {limit}" if limit > 0 else None,
                                          f"OFFSET {offset}" if limit > 0 and offset > 0 else None])))
        return Cursor(self.database.execute(sql, values), columns_, self, query=sql, query_values=values)

    def update(self, query: Selector, new_entry: dict[str, Value]) -> SQLCursor:
        sql, values = selector_to_sql(query) if query else ("", [])
        update_columns: list[str] = [f"{col} = ?" for col in new_entry]
        return self.database.execute(f"UPDATE {self.name} SET {','.join(update_columns)} WHERE {sql}",
                                     [*new_entry.values(), *values])

    def delete(self, query: Selector) -> SQLCursor:
        sql, values = selector_to_sql(query) if query else ("", [])
        return self.database.execute(f"DELETE FROM {self.name} WHERE {sql}", values)

    def add_to_list(self, key: Value, column: str | Column, new_values: Iterable[Value]):
        entry: dict = self._get_exists(key)
        column = column.name if isinstance(column, Column) else self.get_column(column).name
        if changes := any(v not in entry[column] for v in new_values):
            entry[column] = list(entry[column]) + [v for v in new_values if v not in entry[column]]
            self[key] = entry
        return changes

    def remove_from_list(self, key: Value, column: str | Column, new_values: Iterable[Value]):
        entry: dict = self._get_exists(key)
        column = column.name if isinstance(column, Column) else self.get_column(column).name
        if changes := any(v in entry[column] for v in new_values):
            new_list: set = set(entry[column]) - set(new_values)
            entry[column] = sorted(new_list, key=list(entry[column]).index)
            self[key] = entry
        return changes


class UsersTable(Table):
    def save_user(self, user: dict[str, Any], *, replace: bool = False, exist_ok: bool = False):
        self.insert(self.format_entry(user), replace=replace, exists_ok=exist_ok)

    def set_active(self, user: str, active: bool) -> bool:
        if (entry := self._get_exists(user := clean_username(user)))[UsersColumns.ACTIVE.name] is active:
            return False
        self[user] = entry | {UsersColumns.ACTIVE.name: active}
        return True

    def add_folder(self, user: str, folder: str) -> bool:
        return self.add_to_list(clean_username(user), UsersColumns.FOLDERS.value.name, [folder])

    def remove_folder(self, user: str, folder: str) -> bool:
        return self.remove_from_list(clean_username(user), UsersColumns.FOLDERS.value.name, [folder])

    def set_userpage(self, user: str, userpage: str) -> bool:
        entry: dict = self._get_exists(user := clean_username(user))
        if entry[UsersColumns.USERPAGE.value.name] == userpage:
            return False
        entry[UsersColumns.USERPAGE.value.name] = userpage
        self[user] = entry
        return True


class SubmissionsTable(Table):
    @property
    def files_folder(self) -> Path:
        return self.database.settings.files_folder

    def save_submission(self, submission: dict[str, Value | list[Value]], files: list[bytes] = None,
                        thumbnail: bytes = None, *, replace: bool = False, exist_ok: bool = False):
        submission = self.format_entry(submission)
        file_url: list[str] = \
            SubmissionsColumns.FILEURL.value.from_entry(submission[SubmissionsColumns.FILEURL.value.name])

        submission[SubmissionsColumns.FILEEXT.value.name] = SubmissionsColumns.FILEEXT.value.to_entry([
            self.save_submission_file(
                submission[SubmissionsColumns.ID.value.name], file, "submission",
                s[1] if (s := search(r"/[^/]+\.([^.]+)$", file_url[0] if file_url else "")) else "", n)
            for n, file in enumerate(files) if file
        ])
        self.save_submission_thumbnail(submission[SubmissionsColumns.ID.name], thumbnail or None)

        submission[SubmissionsColumns.FILESAVED.value.name] = (
                (0b100 * all(map(bool, files)) * bool(files)) +  # all files were valid
                (0b010 * any(map(bool, files))) +  # at least one file was valid
                (0b001 * bool(thumbnail))  # the thumbnail was valid
        )

        self.insert(submission, replace=replace, exists_ok=exist_ok)

    def save_submission_file(self, submission_id: int, file: bytes | None, name: str, ext: str, n: int = 0,
                             guess_ext: bool = True) -> str:
        if file is None:
            return ""
        assert n >= 0, "n must be zero or positive"

        ext: str = guess_extension(file, ext) if guess_ext else ext
        folder: Path = self.files_folder / tiered_path(submission_id)
        folder.mkdir(parents=True, exist_ok=True)
        folder.joinpath(f"{name}{n if n > 0 else ''}" + f".{ext}" * bool(ext)).write_bytes(file)

        return ext

    def save_submission_thumbnail(self, submission_id: int, file: bytes | None):
        self.save_submission_file(submission_id, file, "thumbnail", "jpg", False)

    def get_submission_files(self, submission_id: int) -> tuple[list[Path] | None, Path | None]:
        if (entry := self[submission_id]) is None or (f := entry[SubmissionsColumns.FILESAVED.value.name]) == 0:
            return None, None
        folder: Path = self.files_folder / tiered_path(submission_id)
        file_ext: list[str] = entry[SubmissionsColumns.FILEEXT.value.name]
        return (
            [folder / f"submission{n or ''}{('.' + ext) if ext else ''}"
             for n, ext in enumerate(file_ext)] if f & 0b10 else None,
            folder / "thumbnail.jpg" if f & 0b01 else None
        )

    def set_filesaved(self, submission_id: int, all_files: bool | int, any_file: bool | int, thumbnail: bool | int):
        filesaved: int = (0b100 * bool(all_files)) + (0b010 * bool(any_file)) + (0b001 * bool(thumbnail))
        if self._get_exists(submission_id)[SubmissionsColumns.FILESAVED.value.name] != filesaved:
            self.update({EQ: {self.key.name: submission_id}}, {SubmissionsColumns.FILESAVED.value.name: filesaved})
            return True
        return False

    def set_folder(self, submission_id: int, folder: str) -> bool:
        if self._get_exists(submission_id)[SubmissionsColumns.FOLDER.value.name] != (folder := folder.lower().strip()):
            self.update({EQ: {self.key.name: submission_id}}, {SubmissionsColumns.FOLDER.value.name: folder})
            return True
        return False

    def set_user_update(self, submission_id: int, update: bool) -> bool:
        if self._get_exists(submission_id)[SubmissionsColumns.USERUPDATE.value.name] != update:
            self.update({EQ: {self.key.name: submission_id}}, {SubmissionsColumns.USERUPDATE.value.name: update})
            return True
        return False

    def add_favorite(self, submission_id: int, user: str) -> bool:
        return self.add_to_list(submission_id, SubmissionsColumns.FAVORITE.value, [clean_username(user)])

    def remove_favorite(self, submission_id: int, user: str) -> bool:
        return self.remove_from_list(submission_id, SubmissionsColumns.FAVORITE.value, [clean_username(user)])

    def add_mention(self, submission_id: int, user: str) -> bool:
        return self.add_to_list(submission_id, SubmissionsColumns.MENTIONS.value, [clean_username(user)])

    def remove_mention(self, submission_id: int, user: str) -> bool:
        return self.remove_from_list(submission_id, SubmissionsColumns.MENTIONS.value, [clean_username(user)])


class JournalsTable(Table):
    def save_journal(self, journal: dict[str, Any], *, replace: bool = False, exist_ok: bool = False):
        self.insert(self.format_entry(journal), replace=replace, exists_ok=exist_ok)

    def set_user_update(self, journal_id: int, update: bool) -> bool:
        if self._get_exists(journal_id)[JournalsColumns.USERUPDATE.value.name] != update:
            self.update({EQ: {self.key.name: journal_id}}, {JournalsColumns.USERUPDATE.value.name: update})
            return True
        return False

    def add_mention(self, journal_id: int, user: str) -> bool:
        return self.add_to_list(journal_id, JournalsColumns.MENTIONS.value, [clean_username(user)])

    def remove_mention(self, journal_id: int, user: str) -> bool:
        return self.remove_from_list(journal_id, JournalsColumns.MENTIONS.value, [clean_username(user)])


class CommentsTable(Table):
    def save_comment(self, comment: dict[str, any], *, replace: bool = False, exist_ok: bool = False):
        self.insert(self.format_entry(comment), replace=replace, exists_ok=exist_ok)

    def get_comments(self, parent_table: str, parent_id: int) -> list[dict]:
        return list(self.select_sql(
            f"{CommentsColumns.PARENT_TABLE.name} = ? and {CommentsColumns.PARENT_ID.name} = ?",
            [parent_table, parent_id],
            order=[f"{CommentsColumns.ID.name} ASC"]))

    def get_comments_tree(self, parent_table: str, parent_id: int) -> list[dict]:
        comments: list[dict] = self.get_comments(parent_table, parent_id)
        return self._make_comments_tree([c for c in comments if c[CommentsColumns.REPLY_TO.name] is None], comments)

    def make_comments_tree(self, comments: list[dict]) -> list[dict]:
        return self._make_comments_tree(comments, comments)

    def _make_comments_tree(self, comments: list[dict], all_comments: list[dict]) -> list[dict]:
        return [com | {"REPLIES": self._make_comments_tree(
            [c for c in all_comments if c[CommentsColumns.REPLY_TO.name] == com[CommentsColumns.ID.name]],
            all_comments)} for com in comments]


class SettingsTable(Table):
    _version_setting: str = "VERSION"
    _files_folder_setting: str = "FILESFOLDER"
    _backup_folder_setting: str = "BACKUPFOLDER"
    _default_files_folder: str = "FA.files"
    _default_backup_folder: str = "FA.backup"

    def __getitem__(self, item: str) -> str | None:
        return (super().__getitem__(item) or {}).get(SettingsColumns.SVALUE.value.name, None)

    def __setitem__(self, key: str, value: str):
        self.insert(self.format_entry({self.key.name: key, SettingsColumns.SVALUE.value.name: value}), replace=True)

    @property
    def version(self):
        return self[self._version_setting]

    @property
    def files_folder(self) -> Path:
        folder: Path = Path(self[self._files_folder_setting])
        return folder if folder.is_absolute() else (self.database.path.parent / folder).resolve()

    @files_folder.setter
    def files_folder(self, value: str | Path):
        self[self._files_folder_setting] = str(value)

    @property
    def backup_folder(self) -> Path | None:
        folder: str | None = self[self._backup_folder_setting]
        if folder is None:
            return None
        return p if (p := Path(folder)).is_absolute() else (self.database.path.parent / p).resolve()

    @backup_folder.setter
    def backup_folder(self, value: str | Path | None):
        if value is None:
            del self[self._backup_folder_setting]
        else:
            self[self._backup_folder_setting] = str(value)

    def create(self, exists_ignore: bool = False):
        super().create(exists_ignore=exists_ignore)
        self.insert({SettingsColumns.SETTING.value.name: self._files_folder_setting,
                     SettingsColumns.SVALUE.value.name: self._default_files_folder}, exists_ok=True)
        self.insert({SettingsColumns.SETTING.value.name: self._version_setting,
                     SettingsColumns.SVALUE.value.name: __version__}, exists_ok=True)


class HistoryTable(Table):
    def __iter__(self) -> Generator[dict[str, Value], None, None]:
        return self.select(order=[self.key.name]).entries

    def add_event(self, event: str, time: datetime = None):
        self[time or datetime.now()] = {HistoryColumns.EVENT.value.name: event}


class Database:
    def __init__(self, path: str | PathLike | Path, *, init: bool = False, check_connections: bool = True,
                 check_version: bool = True, read_only: bool = False, autocommit: bool = False):
        self.path: Path = Path(path).resolve()
        self.read_only: bool = read_only

        if check_connections:
            self.check_connection()

        self.connection: Connection = connect(self.path.as_uri() + ("?mode=ro" if read_only else ""), uri=True)
        self.autocommit = autocommit

        self.users: UsersTable = UsersTable(self, users_table, UsersColumns.as_list())
        self.submissions: SubmissionsTable = SubmissionsTable(self, submissions_table, SubmissionsColumns.as_list())
        self.journals: JournalsTable = JournalsTable(self, journals_table, JournalsColumns.as_list())
        self.comments: CommentsTable = CommentsTable(self, comments_table, CommentsColumns.as_list())
        self.settings: SettingsTable = SettingsTable(self, settings_table, SettingsColumns.as_list())
        self.history: HistoryTable = HistoryTable(self, history_table, HistoryColumns.as_list())

        self.committed_changes: int = self.total_changes

        if self.is_formatted:
            if check_version:
                self.check_version()
        elif init:
            self.init()

    def __getitem__(self, name: str) -> Table:
        return Table(self, name.upper())

    def __enter__(self):
        return self

    def __exit__(self, _exc_type, _exc_val, _exc_tb):
        self.close()

    def __contains__(self, value: str | Table):
        value = value.name.lower() if isinstance(value, Table) else value.lower()
        return value in [t.name.lower() for t in self.tables]

    @property
    def autocommit(self):
        return self.connection.isolation_level is None

    @autocommit.setter
    def autocommit(self, value: bool):
        self.connection.isolation_level = None if value else ""

    @property
    def tables(self) -> list[Table]:
        return [
            Table(self, name)
            for [name] in self.connection.execute(
                """select name from sqlite_master
                where type = 'table'
                and name not like 'sqlite_%'
                order by 1;"""
            )
        ]

    @property
    def total_changes(self) -> int:
        return int(self.connection.total_changes)

    @property
    def is_clean(self) -> bool:
        return self.total_changes == self.committed_changes

    @property
    def is_formatted(self):
        return settings_table in self and self.version is not None

    @property
    def is_open(self) -> bool:
        try:
            self.connection.execute("SELECT * FROM sqlite_master")
            return True
        except ProgrammingError:
            return False

    @property
    def version(self) -> str | None:
        return self.settings.version

    def init(self):
        self.users.create(exists_ignore=True)
        self.submissions.create(exists_ignore=True)
        self.journals.create(exists_ignore=True)
        self.comments.create(exists_ignore=True)
        self.settings.create(exists_ignore=True)
        self.history.create(exists_ignore=True)

    def check_connection(self: Type["Database"] | str | PathLike | Path, raise_for_error: bool = True, limit: int = 0
                         ) -> list[Process]:
        return find_connections(self.path if isinstance(self, Database) else Path(self), raise_for_error, limit)

    def check_version(self, raise_for_error: bool = True) -> VersionError | None:
        err: VersionError | None = compare_version(self.version)
        if raise_for_error and err:
            raise err
        return err

    def execute(self, sql: str, parameters: Iterable = None) -> SQLCursor:
        return self.connection.execute(sql, parameters or [])

    def commit(self):
        self.connection.commit()
        self.committed_changes = self.total_changes

    def rollback(self):
        self.execute("ROLLBACK")

    def reset(self, *, init: bool = False, check_connections: bool = True, check_version: bool = True,
              read_only: bool = None, autocommit: bool = None):
        self.close()
        self.connection = None
        self.__init__(self.path, init=init, check_connections=check_connections, check_version=check_version,
                      read_only=self.read_only if read_only is None else read_only,
                      autocommit=self.autocommit if autocommit is None else autocommit)

    def upgrade(self, *, check_connections: bool = True, read_only: bool = None, autocommit: bool = None):
        self.connection = update_database(self.connection, __version__)
        self.reset(check_connections=check_connections, check_version=False,
                   read_only=self.read_only if read_only is None else read_only,
                   autocommit=self.autocommit if autocommit is None else autocommit)

    def merge(self, db_b: 'Database', *cursors: Cursor, replace: bool = True, exist_ok: bool = True):
        copy_cursors(self, cursors or [db_b.users.select(), db_b.submissions.select(), db_b.journals.select()],
                     replace=replace, exist_ok=exist_ok)

    def copy(self, db_b: 'Database', *cursors: Cursor, replace: bool = True, exist_ok: bool = True):
        copy_cursors(db_b, cursors or [self.users.select(), self.submissions.select(), self.journals.select()],
                     replace=replace, exist_ok=exist_ok)

    def backup(self, *, folder: Path = None, date_format: str = "%Y-%m-%d %H.%M.%S"):
        folder: Path | None = folder or self.settings.backup_folder
        if folder is None:
            raise ValueError("No backup folder set in database settings")
        m_time: datetime = datetime.fromtimestamp(self.path.stat().st_mtime)
        folder.mkdir(parents=True, exist_ok=True)
        backup_file: Path = folder / (f"{self.path.name.removesuffix(self.path.suffix)} "
                                      f"{m_time.strftime(date_format or '%Y-%m-%d %H.%M.%S')}"
                                      f"{self.path.suffix}")
        try:
            copy2(self.path, backup_file.with_suffix(".tmp"))
            move(backup_file.with_suffix(".tmp"), backup_file)
        finally:
            backup_file.with_suffix(".tmp").unlink(missing_ok=True)

    def close(self):
        self.connection.close()



from typing import Optional
from typing import Union

from .exceptions import UnknownSelector
from .types import Value

Selector = dict[str, Union[dict[str, Union[Value, list[Value]]], 'Selector', list['Selector']]]
SELECTOR_NOT = NOT = "$not"
SELECTOR_AND = AND = "$and"
SELECTOR_OR = OR = "$or"
SELECTOR_EQ = EQ = "$eq"
SELECTOR_NE = NE = "$ne"
SELECTOR_GT = GT = "$gt"
SELECTOR_LT = LT = "$lt"
SELECTOR_GE = GE = "$ge"
SELECTOR_LE = LE = "$le"
SELECTOR_IN = IN = "$in"
SELECTOR_INSTR = INSTR = "$instr"
SELECTOR_BETWEEN = BETWEEN = "$between"
SELECTOR_LIKE = LIKE = "$like"
SELECTOR_GLOB = GLOB = "$glob"


def flatten(list_old: list) -> list:
    list_new = []
    for i in list_old:
        list_new.extend(flatten(i)) if isinstance(i, list) else list_new.append(i)
    return list_new


def selector_to_sql(selector: Selector) -> tuple[str, list[Value]]:
    sql, values = "", []
    assert isinstance(selector, dict), "selector needs to be of type dict"
    for key, value in selector.items():
        if key in (AND, OR):
            assert isinstance(value, list) and all(isinstance(v, dict) for v in value)
        elif key in (NOT, EQ, NE, GT, LT, GE, LE, IN, INSTR, BETWEEN, LIKE, GLOB):
            assert isinstance(value, dict)
        else:
            raise UnknownSelector(key)

        if key == SELECTOR_NOT:
            sql_, values_ = selector_to_sql(value)
            sql = f"not ({sql})"
            values.extend(flatten(values_))
        if key == SELECTOR_AND:
            sql_values = list(map(selector_to_sql, value))
            sql = f"({' and '.join([s for s, _ in sql_values])})"
            values.extend(flatten([v for _, v in sql_values]))
        elif key == SELECTOR_OR:
            sql_values = list(map(selector_to_sql, value))
            sql = f"({' or '.join([s for s, _ in sql_values])})"
            values.extend(flatten([v for _, v in sql_values]))
        elif key == SELECTOR_EQ:
            sql = f"{(k := [*value.keys()][0])} = ?"
            values.append(value[k])
        elif key == SELECTOR_NE:
            sql = f"{(k := [*value.keys()][0])} != ?"
            values.append(value[k])
        elif key == SELECTOR_GT:
            sql = f"{(k := [*value.keys()][0])} > ?"
            values.append(value[k])
        elif key == SELECTOR_LT:
            sql = f"{(k := [*value.keys()][0])} < ?"
            values.append(value[k])
        elif key == SELECTOR_GE:
            sql = f"{(k := [*value.keys()][0])} >= ?"
            values.append(value[k])
        elif key == SELECTOR_LE:
            sql = f"{(k := [*value.keys()][0])} <= ?"
            values.append(value[k])
        elif key == SELECTOR_IN:
            sql = f"{(k := [*value.keys()][0])} in " + \
                  f"({','.join(['?'] * len(vs if isinstance(vs := value[k], list) else [vs]))})"
            values.extend(vs)
        elif key == SELECTOR_INSTR:
            sql = f"instr({(k := [*value.keys()][0])}, ?)"
            values.append(value[k])
        elif key == SELECTOR_BETWEEN:
            assert isinstance((v := value[(k := [*value.keys()][0])]), list)
            sql = f"{k} between ? and ?"
            values.extend(v[0:2])
        elif key in (SELECTOR_LIKE, SELECTOR_GLOB):
            assert isinstance((v := value[(k := [*value.keys()][0])]), str)
            sql = f"{k} like ?" if key == SELECTOR_LIKE else f"{k} glob ?"
            values.append(v)

    return sql, values


class SelectorBuilder:
    def __init__(self, field: str = None):
        self.field: Optional[str] = field

    def __sub__(self, other: Selector) -> Selector:  # NOT
        return {SELECTOR_NOT: other}

    def __and__(self, other: list[Selector]) -> Selector:  # AND
        return {SELECTOR_AND: other}

    def __or__(self, other: list[Selector]) -> Selector:  # OR
        return {SELECTOR_OR: other}

    def __eq__(self, value: Value) -> Selector:  # EQ
        return {SELECTOR_EQ: {self.field: value}}

    def __ne__(self, value: Value) -> Selector:  # NE
        return {SELECTOR_NE: {self.field: value}}

    def __gt__(self, value: Value) -> Selector:  # GT
        return {SELECTOR_GT: {self.field: value}}

    def __lt__(self, value: Value) -> Selector:  # LT
        return {SELECTOR_LT: {self.field: value}}

    def __ge__(self, value: Value) -> Selector:  # GE
        return {SELECTOR_GE: {self.field: value}}

    def __le__(self, value: Value) -> Selector:  # LE
        return {SELECTOR_LE: {self.field: value}}

    def __truediv__(self, value: Union[Value, list[Value]]) -> Selector:  # IN
        return {SELECTOR_IN: {self.field: value}}

    def __floordiv__(self, value: Value) -> Selector:  # INSTR
        return {SELECTOR_INSTR: {self.field: value}}

    def __xor__(self, values: list[Value]) -> Selector:  # BETWEEN
        return {SELECTOR_BETWEEN: {self.field: values[:2]}}

    def __mod__(self, value: Value) -> Selector:  # LIKE
        return {SELECTOR_LIKE: {self.field: value}}

    def __mul__(self, value: Value) -> Selector:  # GLOB
        return {SELECTOR_GLOB: {self.field: value}}



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['falocalrepo_database']

package_data = \
{'': ['*']}

install_requires = \
['chardet>=4.0.0,<5.0.0', 'filetype>=1.0.13,<2.0.0', 'psutil>=5.9.1,<6.0.0']

setup_kwargs = {
    'name': 'falocalrepo-database',
    'version': '5.3.3',
    'description': 'Database functionality for falocalrepo.',
    'long_description': '<div align="center">\n\n<img alt="logo" width="400" src="https://raw.githubusercontent.com/FurryCoders/Logos/main/logos/falocalrepo-database-transparent.png">\n\n# FALocalRepo-Database\n\nDatabase functionality for [falocalrepo](https://pypi.org/project/falocalrepo).\n\n[![](https://img.shields.io/github/v/tag/FurryCoders/falocalrepo-database?label=github&sort=date&logo=github&color=blue)](https://github.com/FurryCoders/falocalrepo-database)\n[![](https://img.shields.io/pypi/v/falocalrepo-database?logo=pypi)](https://pypi.org/project/falocalrepo-database/)\n[![](https://img.shields.io/pypi/pyversions/falocalrepo-database?logo=Python)](https://www.python.org)\n\n</div>\n\n## Tables\n\nTo store its information, the database uses separate tables: `USERS`, `SUBMISSIONS`, `JOURNALS`, `SETTINGS`,\nand `HISTORY`.\n\n**Note**: bar-separated lists are formatted as `|item1||item2|` to properly isolate all elements\n\n### Users\n\nThe users\' table contains a list of all the users that have been download with the program, the folders that have been\ndownloaded, and the submissions found in each of those.\n\nEach entry contains the following fields:\n\n* `USERNAME` the URL username of the user (no underscores or spaces)\n* `FOLDERS` the folders downloaded for that specific user, sorted and bar-separated\n* `ACTIVE` `1` if the user is active, `0` if not\n* `USERPAGE` the user\'s profile text\n\n### Submissions\n\nThe submissions\' table contains the metadata of the submissions downloaded by the program and information on their files\n\n* `ID` the id of the submission\n* `AUTHOR` the username of the author (uploader) in full format\n* `TITLE`\n* `DATE` upload date in ISO format _YYYY-MM-DDTHH:MM_\n* `DESCRIPTION` description in html format\n* `TAGS` bar-separated tags\n* `CATEGORY`\n* `SPECIES`\n* `GENDER`\n* `RATING`\n* `TYPE` image, text, music, or flash\n* `FILEURL` a bar-separated list of the remote URLs for the submission files\n* `FILEEXT` a bar-separated list of extensions of the downloaded files. Can be empty if the file contained errors and\n  could not be recognised upon download\n* `FILESAVED` file and thumbnail download status as a 3bit flag:`xx1` if thumbnail was downloaded, `xx0` if not; `x1x`\n  if at least one file was valid `x0x` if not; `1xx` if all given files where valid, `0xx` if not. Possible values\n  are `0` through `7` (3 bit).\n* `FAVORITE` a bar-separated list of users that have "faved" the submission\n* `MENTIONS` a bar-separated list of users that are mentioned in the submission description as links\n* `FOLDER` the folder of the submission (`gallery` or `scraps`)\n* `USERUPDATE` `1` if the submission was added as a user update otherwise `0`\n\n### Journals\n\nThe journals\' table contains the metadata of the journals downloaded by the program.\n\n* `ID` the id of the journal\n* `AUTHOR` the username of the author (uploader) in full format\n* `TITLE`\n* `DATE` upload date in ISO format _YYYY-MM-DDTHH:MM_\n* `CONTENT` content in html format\n* `MENTIONS` a bar-separated list of users that are mentioned in the journal content as links\n* `USERUPDATE` `1` if the journal was added as a user update otherwise `0`\n\n### Comments\n\nThe comments\' table contains the metadata of the journals and submissions stored in the database.\n\n* `ID` the id of the comment\n* `PARENT_TABLE` `SUBMISSIONS` if the comment relates to a submission, `JOURNAL` if the comment relates to a journal\n* `PARENT_ID` the id of the parent object (submission or journal)\n* `REPLY_TO` the id of the parent comment, if the comment is a reply\n* `AUTHOR` the username of the author in full format\n* `DATE` post date in ISO format _YYYY-MM-DDTHH:MM:SS_\n* `TEXT` the text of the comment\n\n### Settings\n\nThe settings table contains settings for the program and variable used by the database handler and main program.\n\n* `COOKIES` cookies for the download program, stored in JSON format\n* `FILESFOLDER` location of downloaded submission files\n* `VERSION` database version\n\n### History\n\nThe history table holds events related to the database.\n\n* `TIME` event time in ISO format _YYYY-MM-DDTHH:MM:SS.ssssss_\n* `EVENT` the event description\n\n## Submission Files\n\nThe `save_submission` functions saves the submission metadata in the database and stores the files.\n\nSubmission files are saved in a tiered tree structure based on their submission ID. IDs are zero-padded to 10 digits and\nthen broken up in 5 segments of 2 digits; each of these segments represents a folder tha will be created in the tree.\n\nFor example, a submission `1457893` will be padded to `0001457893` and divided into `00`, `01`, `45`, `78`, `93`. The\nsubmission file will then be saved as `00/01/45/78/93/submission.file` with the correct extension extracted from the\nfile itself (FurAffinity links do not always contain the right extension and sometimes confuse JPEG and PNG).\n\n## Upgrading Database\n\n_Note:_ versions prior to 4.19.0 are not supported by falocalrepo-database version 5.0.0 and above. To update from\nthose, use [falocalrepo v3.25.0](https://pypi.org/project/falocalrepo/v3.25.0) to upgrade the database to version\n4.19.0.<br/>\n_Note:_ Versions prior to 2.7.0 are not supported by falocalrepo-database version 3.0.0 and above. To update from those\nto the new version use [falocalrepo v2.11.2](https://github.com/FurryCoders/FALocalRepo/releases/tag/v2.11.2) to upgrade\nthe database to version 2.7.0\n',
    'author': 'Matteo Campinoti',
    'author_email': 'matteo.campinoti94@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/FurryCoders/falocalrepo-database',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
